
from django.http import JsonResponse, HttpResponse, HttpResponseNotFound
from django.core.files.storage import default_storage
from .serializers import (
    ImageAssetSerializer, ArticleSerializer, ArticleCollectionSerializer,
    ArticleOrderSerializer, ArticleCollectionOrderSerializer,
    RichTextAssetSerializer, TextAssetSerializer, PasswordChangeSerializer,
    LoginSerializer, PasswordResetRequestSerializer,
    PasswordResetConfirmSerializer, ConfirmEmailSerializer, SignUpSerializer,
    UserSerializer, FeatureOptionSerializer, FeatureUrlSerializer
)
from .models import (
    ImageAsset, Article, ArticleCollection, ArticleOrder,
    ArticleCollectionOrder, RichTextAsset, TextAsset,
    FeatureUrl, FeatureOption
)
from .assets import get_assets
from .auth import generate_verify_email_token
from . import errorcodes, get_signup_serializer, get_user_serializer
from ripple.MailFilter import filtered_send_mail
from rest_framework.exceptions import (
    ValidationError as RestValidationError, APIException,
    PermissionDenied, AuthenticationFailed
)
import coreapi
import functools
import json
import logging
import time
import traceback

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import SetPasswordForm
from django.contrib.auth.tokens import (
    default_token_generator, PasswordResetTokenGenerator)
from django.contrib.auth import (
    authenticate, login as login_, logout as logout_)
from django.contrib.auth.password_validation import validate_password
from django.core import signing
from django.core.exceptions import ValidationError
from django.db import transaction
from django_filters.rest_framework import DjangoFilterBackend
from django.http import HttpResponse
from django.shortcuts import render
from django.template.loader import render_to_string
from django.utils.decorators import method_decorator
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.debug import sensitive_post_parameters

from rest_framework import status, views, viewsets
from rest_framework.decorators import api_view, schema
from rest_framework.permissions import IsAdminUser, IsAuthenticated, AllowAny
from rest_framework.response import Response
from rest_framework.schemas import AutoSchema
from rest_framework.generics import (
    ListAPIView, GenericAPIView, CreateAPIView,
    UpdateAPIView, RetrieveAPIView
)

from django.core.cache import cache

logger = logging.getLogger(__name__)
User = get_user_model()
DAY = 60 * 60 * 24


def handle_errors(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)

        except Exception as e:
            logger.error(traceback.format_exc())

            # let rest_framework handle APIExceptions
            if(isinstance(e, APIException)):
                raise e

            return Response(traceback.format_exc(), status=500)
    return wrapper


unpack_token_schema = AutoSchema(manual_fields=[
    coreapi.Field("token", required=True, location="form",
                  type="string", description="token here"),
])


class FeatureOptionViewSet(viewsets.ModelViewSet):
    permission_classes = [IsAdminUser]
    queryset = FeatureOption.objects.all()
    serializer_class = FeatureOptionSerializer


class FeatureUrlViewSet(viewsets.ModelViewSet):
    permission_classes = [IsAdminUser]
    queryset = FeatureUrl.objects.all()
    serializer_class = FeatureUrlSerializer


@api_view(['GET'])
@schema(unpack_token_schema)
@handle_errors
def unpack_token(request):
    """
    Unpack a signed token, return as JSON data.
    """
    token = request.GET.get('token')
    data = signing.loads(token)
    return Response(data)


def validate_expiry(timestamp, msg=''):
    if int(time.time()) > timestamp + 2 * DAY:
        raise ValueError(
            "Sorry, your session has expired, please try %s again.")


@api_view(['POST'])
@handle_errors
def logout(request):
    """
    Log the current user out.

    Returns no content
    """
    logout_(request)
    return Response(None, status.HTTP_204_NO_CONTENT)


@api_view(['GET'])
def get_session_data(request):
    ret = {}

    if 'API' in request.session.keys():
        ret = request.session['API']

    return Response(ret)


@api_view(['POST', 'PATCH'])
def update_session_data(request):
    if 'readonly' in request.data:
        del request.data['readonly']
    if request.method == 'POST':
        request.session['API'] = request.data
    elif request.method == 'PATCH':
        if 'API' not in request.session.keys():
            request.session['API'] = {}
        request.session['API'] = {**request.session['API'], **request.data}

    return Response(None, status=status.HTTP_204_NO_CONTENT)


def session(request):
    ret = {}

    if 'API' in request.session.keys():
        ret = request.session['API']

    return HttpResponse("Session = "+json.dumps(ret),
                        content_type="text/javascript")


def simple_upload(request):
    if request.method == 'POST' and request.FILES['upload'] and request.user.is_staff:
        myfile = request.FILES['upload']
        filename = default_storage.save(myfile.name, myfile)
        uploaded_file_url = default_storage.url(filename)
        return JsonResponse({
            'url': uploaded_file_url
        })
    return None


class TextAssetViewSet(viewsets.ModelViewSet):
    permission_classes = [IsAdminUser]
    queryset = TextAsset.objects.all()
    serializer_class = TextAssetSerializer

    filter_backends = (DjangoFilterBackend,)
    filter_fields = {'module': ['exact'], 'version': [
        'gt'], 'owner': ['exact'], 'published': ['exact'], 'assetid': ['exact']}


class RichTextAssetViewSet(viewsets.ModelViewSet):
    permission_classes = [IsAdminUser]
    queryset = RichTextAsset.objects.all()
    serializer_class = RichTextAssetSerializer
    filter_backends = (DjangoFilterBackend,)
    filter_fields = {'module': ['exact'], 'version': [
        'gt'], 'owner': ['exact'], 'published': ['exact'], 'assetid': ['exact']}


class ImageAssetViewSet(viewsets.ModelViewSet):
    permission_classes = [IsAdminUser]
    queryset = ImageAsset.objects.all()
    serializer_class = ImageAssetSerializer

    filter_backends = (DjangoFilterBackend,)
    filter_fields = {'module': ['exact'], 'version': [
        'gt'], 'owner': ['exact'], 'published': ['exact'], 'assetid': ['exact']}


class ArticleViewSet(viewsets.ModelViewSet):
    permission_classes = [IsAdminUser]
    queryset = Article.objects.all()
    serializer_class = ArticleSerializer

    filter_backends = (DjangoFilterBackend, )
    filter_fields = {'module': ['exact'], 'version': [
        'gt'], 'owner': ['exact'], 'published': ['exact'], 'assetid': ['exact']}


class ArticleCollectionViewSet(viewsets.ModelViewSet):
    permission_classes = [IsAdminUser]
    queryset = ArticleCollection.objects.all()
    serializer_class = ArticleCollectionSerializer

    filter_backends = (DjangoFilterBackend, )
    filter_fields = {'module': ['exact'], 'version': [
        'gt'], 'owner': ['exact'], 'published': ['exact'], 'assetid': ['exact']}


class ArticleOrderViewSet(viewsets.ModelViewSet):
    permission_classes = [IsAdminUser]
    queryset = ArticleOrder.objects.all()
    serializer_class = ArticleOrderSerializer


class ArticleCollectionOrderViewSet(viewsets.ModelViewSet):
    permission_classes = [IsAdminUser]
    queryset = ArticleCollectionOrder.objects.all()
    serializer_class = ArticleCollectionOrderSerializer


def assets(request):
    module = request.GET.get('module', 'main')
    key = 'assets:>'+module
    ret = cache.get(key)

    if ret is None:
        ret = "GlobalAssets = " + \
            json.dumps(get_assets(module=module, user=request.user))
        cache.set(key, ret, 3600)

    return HttpResponse(ret, content_type="text/javascript")


def assets_admin(request):
    module = request.GET.get('module', 'main')
    return HttpResponse("GlobalAssets = "+json.dumps(get_assets(module=module, user=request.user)),
                        content_type="text/javascript")


def get_empty_user_status():
    return {
        'is_authenticated': False,
        'is_admin': False,
        'is_staff': False,
        'is_email_verified': False
    }


class AccountStatusView(views.APIView):

    def get(self, request, *args, **kwargs):

        user = request.user

        if user and user.is_anonymous:
            user = None
        else:
            serialized_user = get_user_serializer()(user)

        response_data = (
            get_empty_user_status()
            if user is None
            else serialized_user.data
        )

        return Response(response_data)


class SignUpView(CreateAPIView):

    serializer_class = get_signup_serializer()

    def perform_create(self, serializer):
        # log in the user after sign up for simplicity
        super().perform_create(serializer)
        user = serializer.instance
        login_(self.request, user)

    def get_serializer_context(self):

        context = super().get_serializer_context()
        context['send_welcome_email'] = True
        return context


class LoginView(GenericAPIView):

    # @method_decorator(sensitive_post_parameters())
    def post(self, request):

        self.request = request

        serializer = LoginSerializer(
            data=request.data, context={'request': request})

        serializer.is_valid(raise_exception=True)

        login_(self.request, serializer.validated_data['user'])

        user_status = get_user_serializer()(request.user)

        return Response(user_status.data, status=status.HTTP_200_OK)


class LogoutView(views.APIView):

    permission_classes = (IsAuthenticated,)

    @handle_errors
    def post(self, request, *args, **kwargs):

        logout_(request)
        return Response({'detail': _('Successfully logged out.')},
                        status=status.HTTP_204_NO_CONTENT)


class PasswordChangeView(GenericAPIView):

    serializer_class = PasswordChangeSerializer
    permission_classes = (IsAuthenticated,)

    # prevents passwords being logged out
    @method_decorator(sensitive_post_parameters(
        'old_password', 'new_password1', 'new_password2'
    ))
    def dispatch(self, request, *args, **kwargs):

        return super(PasswordChangeView, self).dispatch(request, args, **kwargs)

    def post(self, request, *args, **kwargs):

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        serializer.save()

        return Response({'detail': _("Password has been changed")})


class PasswordResetRequestView(GenericAPIView):

    serializer_class = PasswordResetRequestSerializer
    permission_classes = (AllowAny,)

    def post(self, request, *args, **kwargs):

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        serializer.save()
        # TODO red - wording correct?
        return Response({'detail': _((f"An email has been sent to {serializer.data['email']} with"
                                      " instructions to reset your password"))})
        # {"detail": _("Password reset request received.")})


class PasswordResetConfirmView(GenericAPIView):

    serializer_class = PasswordResetConfirmSerializer
    permission_classes = (AllowAny,)

    @method_decorator(sensitive_post_parameters(
        'new_password1', 'new_password2')
    )
    def dispatch(self, *args, **kwargs):
        return super(PasswordResetConfirmView, self).dispatch(*args, **kwargs)

    def post(self, request, *args, **kwargs):

        serializer = self.get_serializer(
            data=request.data)
        serializer.is_valid(raise_exception=True)

        serializer.save()

        serializer.user.send_password_changed_confirmed_email()

        login_(request, serializer.user)
        return Response({'detail': _('Password has been updated.')})


class SendEmailVerificationView(views.APIView):

    permission_classes = (IsAuthenticated,)

    def get(self, request):

        request.user.send_verification_email(request)
        # TODO red - could save token in db and expire old tokens if multiple
        #  emails are requested

        return Response({'detail': 'Verification Email Sent.'})


class ConfirmEmailView(GenericAPIView):

    serializer_class = ConfirmEmailSerializer

    def post(self, request):

        serializer = self.get_serializer(data=request.data)

        serializer.is_valid(raise_exception=True)

        user = serializer.save()

        login_(request, user)

        return Response({'detail': 'Email Verified Successfully'})


class UserUpdateView(UpdateAPIView, RetrieveAPIView):
    """
        get/put/patch end point for updating current
        user model data
    """
    permission_classes = [IsAuthenticated]
    serializer_class = UserSerializer

    def get_queryset(self):

        queryset = get_user_model().objects.filter(
            id=self.request.user.id
        )
        return queryset

    def get_serializer_context(self):
        return {'request': self.request}



# Factories allow test objects to be created much more easily than
# fixtures. Rather than defining the data to be created, factories
# define a blueprint for creating objects which can be called multiple
# times

# factory_boy

from typing import Any, Sequence

from django.contrib.auth import get_user_model
from factory import DjangoModelFactory, Faker, post_generation, LazyAttribute


class UserFactory(DjangoModelFactory):

    first_name = Faker("first_name")
    last_name = Faker("last_name")
    # generate the email address from the random username
    phone_number = '01234567890'
    email = LazyAttribute(
        lambda obj: f'{obj.first_name.lower()}.{obj.last_name.lower()}@rippleenergy.com')

    @post_generation
    def password(self, create: bool, extracted: Sequence[Any], **kwargs):

        # randomly generate password is one isn't passed in
        password = extracted if extracted is not None else Faker(
            "password",
            length=42,
            special_chars=True,
            digits=True,
            upper_case=True,
            lower_case=True,
        ).generate(extra_kwargs={})

        # self.password_confirm = password
        if isinstance(self, dict):
            self['password_confirm'] = password
        else:
            self.password_confirm = password

        if create:

            self.set_password(password)

    class Meta:
        model = get_user_model()
        # https://factoryboy.readthedocs.io/en/latest/orms.html#factory.django.DjangoOptions.django_get_or_create
        django_get_or_create = ["email"]
        exclude = ('password_confirm',)



# Generated by Django 2.1 on 2019-10-16 14:55

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ripple', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='user',
            name='phone_number',
            field=models.CharField(max_length=20, null=True, validators=[django.core.validators.RegexValidator(message='Phone number format must be entered in the format: (+)9999999 with optional international code e.g (+4499999). Between 8 and 15 digits allowed.', regex='^\\+?1?\\d{9,15}$')]),
        ),
    ]



LOCAL_SETTINGS = {}


class Settings:
    SETTINGS = LOCAL_SETTINGS

    @staticmethod
    def get(name, default=None):
        if name in Settings.SETTINGS.keys():
            return Settings.SETTINGS[name]
        if default is not None:
            return default
        raise Exception(('Setting %s not found. Add %s to config/LOCAL_SETTINGS.py') % (name, name))



import phonenumbers
from phonenumbers.phonenumberutil import NumberParseException

from django.core.exceptions import ValidationError


def validate_gb_phone_number(value):
    try:

        parsed = phonenumbers.parse(value, 'GB')
        valid = phonenumbers.is_valid_number(parsed)
    except NumberParseException:
        valid = False
    finally:
        if not valid:
            raise ValidationError("Phone number is not a valid UK number.")




from .models import ImageAsset, Article, ArticleCollection, ArticleOrder
from .models import ArticleCollectionOrder, RichTextAsset, TextAsset
from .serializers import RichTextAssetSerializerB, ImageAssetSerializerB, TextAssetSerializerB
from .serializers import ArticleSerializerB, ArticleCollectionSerializerB
from django.db.models import Max, F, OuterRef, Subquery, Q

import datetime
from django.utils import timezone


def get_assets(module, version=None, user=None, update=False):

    if not version:
        version=timezone.now().strftime("%Y-%m-%dT%H:%M:%S%fZ")

    version_dt = datetime.datetime.strptime(version, "%Y-%m-%dT%H:%M:%S%fZ")
    version_dt = timezone.make_aware(version_dt)

    filter = []

    filter.append(Q(module=module))

    if user and (user.is_superuser or user.is_staff) :
        filter.append(Q(owner=user.id) | Q(published=True))
    else:
        filter.append(Q(published=True))

    assets = {'version': version, 'module': module,
              'is_admin': user and user.is_superuser,
              'userid': (user and user.id) or -1}

    #images

    latest = ImageAsset.objects.filter(*filter).filter(assetid=OuterRef('assetid')).order_by('-updated_at')
    os = ImageAsset.objects.filter(*filter).filter(id=Subquery(latest.values('id')[:1]))
    s = ImageAssetSerializerB(os, many=True)
    assets['images'] = s.data

    # #text

    latest = TextAsset.objects.filter(*filter).filter(assetid=OuterRef('assetid')).order_by('-updated_at')
    os = TextAsset.objects.filter(*filter).filter(id=Subquery(latest.values('id')[:1]))
    s = TextAssetSerializerB(os, many=True)
    assets['text'] = s.data

    #richtext

    latest = RichTextAsset.objects.filter(*filter).filter(assetid=OuterRef('assetid')).order_by('-updated_at')
    os = RichTextAsset.objects.filter(*filter).filter(id=Subquery(latest.values('id')[:1]))
    s = RichTextAssetSerializerB(os, many=True)
    assets['richtext'] = s.data

    #articles

    os = Article.objects.filter(*filter)
    s = ArticleSerializerB(os, many=True)
    assets['articles'] = s.data

    #article collections
    # this will need to be in order

    article_collections = []

    os = ArticleCollection.objects.filter(*filter)
    for a in os:
        s = ArticleCollectionSerializerB(a)
        ac = s.data

        items = []

        a_orders = {}
        ac_orders = {}

        for ao in a.articles.filter(*filter):
            orders = ArticleOrder.objects.filter(article=ao.id, article_collection=a.id)
            order = orders[0].order
            key = "%d:%d" % (ao.id, a.id)
            if len(orders) > 0:
                if key not in a_orders:
                    a_orders[key] = 0
                order = orders[a_orders[key]].order
                a_orders[key] += 1

            items.append({'order': order, 'type': 'article', 'assetid': ao.assetid, 'id': ao.id})

        for aci in a.article_collections.filter(*filter):
            orders = ArticleCollectionOrder.objects.filter(container_article_collection=a.id, article_collection=aci.id)
            order = orders[0].order
            key = "%d:%d" % (a.id, aci.id)
            if len(orders) > 0:
                if key not in ac_orders:
                    ac_orders[key] = 0
                order = orders[ac_orders[key]].order
                ac_orders[key] += 1

            items.append({'order': order, 'type': 'article_collection', 'assetid': aci.assetid, 'id': aci.id})

        items.sort(key=lambda a: a['order'])
        ac['items'] = items
        article_collections.append(ac)

    assets['article_collections'] = article_collections

    return assets





from django.conf import settings
from django.core import mail
from django.core.mail import send_mail, backends
from django.core.mail import EmailMultiAlternatives
import re
import logging

logger = logging.getLogger(__name__)

ALLOW_EXP = settings.EMAIL_ALLOW_EXP and [
    re.compile(e) for e in settings.EMAIL_ALLOW_EXP.split(' ')]
BLOCK_EXP = settings.EMAIL_BLOCK_EXP and [
    re.compile(e) for e in settings.EMAIL_BLOCK_EXP.split(' ')]


def filter_email(from_email, recipient_list):

    recp1 = []
    recp2 = []

    if settings.EMAIL_FORCE_SENDER:
        from_email = settings.EMAIL_FORCE_SENDER

    if from_email is None or from_email == "":
        from_email = settings.EMAIL_DEFAULT_SENDER
    if ALLOW_EXP and len(ALLOW_EXP) > 0:
        for r in recipient_list:
            append = False
            for e in ALLOW_EXP:
                if e and e.match(r):
                    append = True
            if append:

                recp1.append(r)
            else:
                logger.info('%s filtered by ALLOW_EXP: %s' %
                            (r, settings.EMAIL_ALLOW_EXP))

    else:
        recp1 = recipient_list

    if BLOCK_EXP and len(BLOCK_EXP) > 0:
        for r in recp1:
            block = False

            for e in BLOCK_EXP:
                if e and e.match(r):
                    block = True
            if not block:
                recp2.append(r)
            else:
                logger.info('%s blocked by BLOCK_EXP: %s' %
                            (r, settings.EMAIL_BLOCK_EXP))

    else:
        recp2 = recp1

    return from_email, recp2


def filtered_send_mail(
    subject,
    text_message,
    from_email,
    recipient_list,
    fail_silently=False,
    auth_user=None,
    auth_password=None,
    connection=None,
    html_message=None,
    sg_template_id=None,
    sg_context=None,
    sg_substitutions=None
):
    # TODO red attachments
    from_email, recp2 = filter_email(from_email, recipient_list)
    emails = []
    for to_email in recp2:
        # https://docs.djangoproject.com/en/2.2/topics/email/#sending-alternative-content-types
        email = EmailMultiAlternatives(
            subject, text_message, from_email, [to_email])

        if sg_template_id:
            email.template_id = sg_template_id
            if html_message is None:
                html_message = '_'  # required to get sendgrid emails sent as html, not texts

        if html_message:
            email.attach_alternative(html_message, 'text/html')

        if sg_context:
            email.dynamic_template_data = sg_context

        if sg_substitutions:
            email.substitutions = sg_substitutions

        emails.append(email)
    if len(emails):

        connection = mail.get_connection(fail_silently=fail_silently)
        connection.send_messages(emails)

        return emails



from datetime import timedelta
import re
from django.conf import settings

from django.core import signing
from django.core.signing import TimestampSigner
from django.http import HttpResponse, HttpResponseNotFound

from .models import FeatureOption
from django.core.cache import cache

url_re = re.compile(settings.PRIVATE_URL_RE)
exceptions_re = settings.PRIVATE_URL_EXCEPTION_RE and re.compile(settings.PRIVATE_URL_EXCEPTION_RE)


class TesterMiddleware:

    def __init__(self, get_response):
        self.get_response = get_response

        # One-time configuration and initialization.
        # raise MiddlewareNotUsed

    def __call__(self, request):

        # Code to be executed for each request before
        # the view (and later middleware) are called.
        passkey = request.GET.get('pass', None )
        signer = TimestampSigner()
        request.session['features'] = []
        if not (request.path.startswith('/admin/')
                or request.path in ['/healthcheck', '/robots.txt', '/troubleshoot', '/account/verify']
                or (exceptions_re and exceptions_re.match(request.path))
                or passkey in settings.PRIVATE_URL_ACCESS_KEYS
                ):

            fo = cache.get('featureoptions')

            if fo is None:
                fo = [{'name': f.name, 'urls_re': [re.compile(u.regex) for u in f.urls.all()]}
                      for f in FeatureOption.objects.filter(active=True).all()]
                cache.set('featureoptions', fo, 3600)

            if settings.PRIVATE_URL_RE:
                fo.append({'name': 'site-settings', 'urls_re': [url_re]})

            for f in fo:
                block = True
                if f['name'] in request.COOKIES:
                    value = request.COOKIES[f['name']]
                    try:
                        value = signer.unsign(value, max_age=timedelta(days=2))
                        block = False
                        request.session['features'].append(f['name'])
                    except signing.BadSignature:
                        pass

                if 'staff' in request.COOKIES:
                    try:
                        signer.unsign(
                            request.COOKIES['staff'], max_age=timedelta(days=7))
                        block = False
                    except signing.BadSignature:
                        pass

                for key in settings.PRIVATE_URL_ACCESS_KEYS:
                    if key in request.COOKIES:
                        try:
                            signer.unsign(
                                request.COOKIES[key], max_age=timedelta(hours=settings.PRIVATE_URL_ACCESS_TIME))
                            block = False
                        except signing.BadSignature:
                            pass

                if block and not request.user.is_staff:
                    for u in f['urls_re']:
                        if u.match(request.path):
                            return HttpResponse(f'<h1>Not Found {request.path}</h1>')

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        if hasattr(request, 'user') and request.user.is_staff:
            response.set_cookie('staff', signer.sign('staff'), max_age=86400*7)
        if passkey in settings.PRIVATE_URL_ACCESS_KEYS:
            response.set_cookie(passkey, signer.sign(passkey), max_age=settings.PRIVATE_URL_ACCESS_TIME*3600)

        return response



from django.db import models
from django.conf import settings
from django.contrib.auth.base_user import AbstractBaseUser, BaseUserManager
from django.contrib.auth.models import PermissionsMixin
from django.contrib.auth.tokens import default_token_generator
from django.core.mail import send_mail
from django.core.validators import RegexValidator
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.encoding import force_bytes, force_text
from django.utils.http import urlsafe_base64_encode, urlsafe_base64_decode
from django.utils.translation import ugettext_lazy as _


from .auth import generate_verify_email_token
from .validators import validate_gb_phone_number
from .MailFilter import filtered_send_mail
from django.core import signing
from datetime import timedelta

# remove username from the default django user model. (Creating a custom user model based on the original django model)
# email is used as the unique account 'username'
import logging
logger = logging.getLogger(__name__)


class UserManager(BaseUserManager):
    use_in_migrations = True

    def _create_user(self, email, password,
                     is_staff, is_superuser, **extra_fields):
        """
        Creates and saves a User with the given username, email and password.
        """
        now = timezone.now()
        email = self.normalize_email(email)
        user = self.model(email=email,
                          is_staff=is_staff, is_active=True,
                          is_superuser=is_superuser,
                          date_joined=now,
                          is_email_verified=False, **extra_fields)
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, email, password=None, **extra_fields):
        return self._create_user(email, password, False, False,
                                 **extra_fields)

    def create_superuser(self, email, password, **extra_fields):
        return self._create_user(email, password, True, True,
                                 **extra_fields)


class User(AbstractBaseUser, PermissionsMixin):

    first_name = models.CharField(_('first name'), max_length=30, blank=True)
    last_name = models.CharField(_('last name'), max_length=30, blank=True)
    email = models.EmailField(_('email address'), unique=True)
    phone_number = models.CharField(
        null=True,
        max_length=20,
        validators=[validate_gb_phone_number]
    )
    profile_pic = models.ImageField(blank=True, null=True)
    is_staff = models.BooleanField(_('staff status'), default=False,
                                   help_text=_('Designates whether the user can log into this admin '
                                               'site.'))
    is_active = models.BooleanField(_('active'), default=True,
                                    help_text=_('Designates whether this user should be treated as '
                                                'active. Unselect this instead of deleting accounts.'))
    date_joined = models.DateTimeField(_('date joined'), default=timezone.now)

    is_email_verified = models.BooleanField(_('email verified'), default=False,
                                            help_text=_('Designates whether email has been verified '))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    objects = UserManager()

    USERNAME_FIELD = 'email'
    REQUIRED_FIELDS = []

    class Meta:
        verbose_name = _('user')
        verbose_name_plural = _('users')

    def get_full_name(self):
        """
        Returns the first_name plus the last_name, with a space in between.
        """
        full_name = '%s %s' % (self.first_name, self.last_name)
        return full_name.strip()

    def get_short_name(self):
        "Returns the short name for the user."
        return self.first_name

    def email_user(self, subject, message, from_email=None, **kwargs):
        """
        Sends an email to this User.
        """
        send_mail(subject, message, from_email, [self.email], **kwargs)

    def _get_base_sg_context(self):
        return {
            'first_name': self.first_name,
            'last_name': self.last_name,
            'email': self.email,
        }

    def send_verification_email(self, request=None):

        preview = ''
        passkey = ''

        if request:
            if settings.ENABLE_PREVIEW and request.session.get('preview', None) is not None:
                preview = '&preview'

            signer = signing.TimestampSigner()

            for key in settings.PRIVATE_URL_ACCESS_KEYS:
                if key in request.COOKIES:
                    try:
                        signer.unsign(
                            request.COOKIES[key], max_age=timedelta(hours=settings.PRIVATE_URL_ACCESS_TIME))
                        passkey = f'&pass={key}'
                    except signing.BadSignature:
                        pass

        token = generate_verify_email_token(self.id, self.email)

        sg_context = {
            'first_name': self.first_name,
            'email': self.email,
            'last_name': self.last_name,
            'verify_url': f"{settings.SITE_URL}/account/verify?token={token}{preview}{passkey}"
        }

        return filtered_send_mail(
            '_',
            '_',
            '',
            [self.email],
            html_message='_',
            sg_template_id=settings.SG_TL_EMAIL_VERIFICATION,
            sg_context=sg_context
        )

    def send_password_reset_email(self):
        uidb64 = urlsafe_base64_encode(force_bytes(self.pk))
        
        token = default_token_generator.make_token(self)

        url = f"{settings.SITE_URL}/account/reset-password-confirm"
        url += f"?uidb64={uidb64}&token={token}"

        sg_context = {
            'first_name': self.first_name,
            'url': url
        }

        filtered_send_mail(
            '_',
            '_',
            'support@rippleenergy.com',
            [self.email],
            html_message='_',
            sg_template_id=settings.SG_TL_RESET_PASSWORD,
            sg_context=sg_context
        )

    def send_password_changed_confirmed_email(self):
        # sends a confirmation to someone that their password has been changed
        # TODO red

        sg_context = {
            **self._get_base_sg_context()
        }
       
        return filtered_send_mail(
            '_',
            '_',
            '',
            [self.email],
            html_message='_',
            sg_template_id=settings.SG_TL_PASSWORD_CHANGED_CONFIRMED,
            sg_context=sg_context
        )

    def send_email_changed_email(self, old_email_address):
        """
        Sends email to a user's old email address 
        to notify that their email has been changed.
        """
        return
        sg_context = {
            'old_email': old_email_address,
            **self._get_base_sg_context()
        }

        return filtered_send_mail(
            '_',
            '_',
            '',
            [old_email_address],
            html_message='_',
            sg_template_id=settings.SG_TL_EMAIL_CHANGED_NOTIFY,
            sg_context=sg_context
        )


class RippleBase(models.Model):

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class AssetBase(RippleBase):

    class Meta:
        abstract = True

    assetid = models.CharField(max_length=200, db_index=True)
    module = models.CharField(
        max_length=200, blank=True, default="", db_index=True)
    version = models.DateTimeField(auto_now=True)
    published = models.BooleanField(default=False)
    owner = models.ForeignKey(
        User, on_delete=models.CASCADE, blank=True, null=True)

    def __str__(self):
        return self.assetid


class TextAsset(AssetBase):

    text = models.CharField(max_length=1000, blank=True, default="")


class RichTextAsset(AssetBase):

    text = models.TextField(blank=True, default="")


class ImageAsset(AssetBase):

    target = models.CharField(max_length=200, default="", db_index=True)
    image = models.ImageField()


class Article(AssetBase):

    layout = models.CharField(
        max_length=200, blank=True, default="", db_index=True)
    image = models.ForeignKey(ImageAsset, on_delete=models.CASCADE)
    text = models.ForeignKey(RichTextAsset, on_delete=models.CASCADE)
    title = models.ForeignKey(TextAsset, on_delete=models.CASCADE)


class ArticleCollection(AssetBase):

    layout = models.CharField(
        max_length=200, blank=True, default="", db_index=True)

    image = models.ForeignKey(ImageAsset, on_delete=models.CASCADE)
    text = models.ForeignKey(RichTextAsset, on_delete=models.CASCADE)
    title = models.ForeignKey(TextAsset, on_delete=models.CASCADE)

    articles = models.ManyToManyField(Article, through='ArticleOrder')
    article_collections = models.ManyToManyField('ArticleCollection', through='ArticleCollectionOrder',
                                                 through_fields=('container_article_collection', 'article_collection'))


class ArticleOrder(models.Model):

    article_collection = models.ForeignKey(
        ArticleCollection, on_delete=models.CASCADE)
    article = models.ForeignKey(Article, on_delete=models.CASCADE)
    order = models.IntegerField()


class ArticleCollectionOrder(models.Model):

    container_article_collection = models.ForeignKey(
        ArticleCollection, related_name="container_articles", on_delete=models.CASCADE)
    article_collection = models.ForeignKey(
        ArticleCollection, related_name="contained_articles", on_delete=models.CASCADE)
    order = models.IntegerField()


class FeatureOption(RippleBase):

    name = models.CharField(max_length=200)
    description = models.CharField(max_length=200, blank=True, default='')
    active = models.BooleanField(default=False)

    def __str__(self):
        return self.name


class FeatureUrl(RippleBase):

    description = models.CharField(max_length=200, blank=True, default='')
    feature = models.ForeignKey(
        FeatureOption, related_name='urls', on_delete=models.CASCADE)
    regex = models.TextField(max_length=800)

    def __str__(self):
        return "{} {}".format(self.feature.name, self.regex[:20])



# Generated by Django 3.0 on 2020-04-02 13:11

from django.db import migrations, models
import ripple.validators


class Migration(migrations.Migration):

    dependencies = [
        ('ripple', '0003_featureoption_featureurl'),
    ]

    operations = [
        migrations.AlterField(
            model_name='user',
            name='phone_number',
            field=models.CharField(max_length=20, null=True, validators=[ripple.validators.validate_gb_phone_number]),
        ),
    ]
